%%
%% %CopyrightBegin%
%%
%% Copyright Ericsson AB 2017. All Rights Reserved
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied
%% See the License for the specific language governing permissions and
%% limitations under the License
%%
%% %CopyrightEnd%
%%
-module(wx_oc_object).
-include_lib("wx/include/wx.hrl").

%%-behaviour(wx_object).  %% commented out avoid warnings

%% wx_object callbacks
-export([init/1]).

-record(state, {}).

init([]) ->
    Init = fun() ->
        Frame0 = wxFrame:new(wx:null(), ?wxID_ANY, "Test wx_object", [{size, {500, 400}}]),
        Frame = wx_object:set_pid(Frame0, self()),
        Sz = wxBoxSizer:new(?wxHORIZONTAL),
        Panel = wxPanel:new(Frame),
        wxSizer:add(Sz, Panel, [{flag, ?wxEXPAND}, {proportion, 1}]),
        wxWindow:show(Frame),
        {Frame, {Frame, Panel}}
    end,
    {Obj, _UserState} = Init(),
    {Obj, #state{}};
init([Init]) ->
    {Obj, _UserState} = Init(),
    {Obj, #state{}}.
